# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import base64
from pathlib import Path
from typing import IO
from urllib.parse import urlparse
from urllib.request import urlopen

import cv2
import numpy as np
import pyarrow as pa
from IPython.core.display import Image as IPyImage
from PIL import Image as PILImage

from pixano.transforms import binary_to_url


class Image:
    """Image type using URI or bytes

    Attributes:
        _uri (str): Image URI
        _bytes (bytes): Image bytes
        _preview_bytes (bytes): Image preview bytes
        uri_prefix (str): URI prefix for relative URIs
    """

    def __init__(
        self,
        uri: str,
        bytes: bytes = None,
        preview_bytes: bytes = None,
        uri_prefix: str = None,
    ):
        """Initialize image from URI or bytes

        Attributes:
            uri (str): Image URI
            bytes (bytes, optional): Image bytes. Defaults to None.
            preview_bytes (bytes, optional): Image preview bytes. Defaults to None.
            uri_prefix (str, optional): URI prefix for relative URIs. Defaults to None.
        """

        self._uri = uri
        self._bytes = bytes
        self._preview_bytes = preview_bytes
        self.uri_prefix = uri_prefix

    @property
    def bytes(self) -> bytes:
        """Return image bytes

        Returns:
            bytes: Image bytes
        """

        if self._bytes is not None:
            return self._bytes
        elif self._uri is not None:
            with self.open() as f:
                return f.read()
        else:
            return None

    @property
    def preview_bytes(self) -> bytes:
        """Return image preview bytes

        Returns:
            bytes: Image bytes
        """

        return self._preview_bytes

    @property
    def url(self) -> str:
        """Return image URL

        Returns:
            str: Image URL
        """

        return binary_to_url(self.bytes)

    @property
    def preview_url(self) -> str:
        """Return image preview URL

        Returns:
            str: Image preview URL
        """

        return binary_to_url(self._preview_bytes)

    @property
    def uri(self) -> str:
        """Return image URI

        Returns:
            str: Image URI
        """

        # Relative URI
        if urlparse(self._uri).scheme == "":
            # If URI prefix exists
            if self.uri_prefix is not None:
                parsed_uri = urlparse(self.uri_prefix)
                if parsed_uri.scheme == "":
                    raise Exception(
                        "URI prefix is incomplete, no scheme provided (http://, file://, ...)"
                    )
                combined_path = Path(parsed_uri.path) / self._uri
                parsed_uri = parsed_uri._replace(path=str(combined_path))
                return parsed_uri.geturl()
            # No URI prefix
            else:
                raise Exception(
                    "Cannot create URI from a relative path without a URI prefix."
                )
        # Complete URI
        else:
            return self._uri

    @property
    def size(self) -> list[int]:
        """Return image size

        Returns:
            list[int]: Image size
        """
        return self.as_pillow().size

    def open(self) -> IO:
        """Open image

        Returns:
            IO: Opened image
        """

        return urlopen(self.uri)

    def as_pillow(self) -> PILImage.Image:
        """Open image as Pillow

        Returns:
            PIL.Image.Image: Image as Pillow
        """

        return PILImage.open(self.open()).convert("RGB")

    def as_cv2(self) -> np.ndarray:
        """Open image as OpenCV

        Returns:
            np.ndarray: Image as OpenCV
        """

        im_arr = np.frombuffer(self.open().read(), dtype=np.uint8)
        return cv2.imdecode(im_arr, cv2.IMREAD_COLOR)

    def display(self, preview=False) -> IPyImage:
        """Display image

        Args:
            preview (bool, optional): True to display image preview instead of full image. Defaults to False.

        Returns:
            IPython.core.display.Image: Image as IPython Display
        """

        im_bytes = self._preview_bytes if preview else self.bytes
        return IPyImage(url=binary_to_url(im_bytes), format=IPyImage(im_bytes).format)

    def to_dict(self) -> dict:
        """Return image attributes as dict

        Returns:
            dict: Image attributes
        """

        return {
            "uri": self._uri,
            "bytes": self._bytes,
            "preview_bytes": self._preview_bytes,
        }


class ImageType(pa.ExtensionType):
    """Externalized image type containing the URI string in UTF-8"""

    def __init__(self):
        super(ImageType, self).__init__(
            pa.struct(
                [
                    pa.field("uri", pa.utf8()),
                    pa.field("bytes", pa.binary()),
                    pa.field("preview_bytes", pa.binary()),
                ]
            ),
            "Image",
        )

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return ImageType()

    def __arrow_ext_scalar_class__(self):
        return ImageScalar


class ImageScalar(pa.ExtensionScalar):
    def as_py(self, uri_prefix: str = None) -> Image:
        return Image(
            self.value["uri"].as_py(),
            self.value["bytes"].as_py(),
            self.value["preview_bytes"].as_py(),
            uri_prefix,
        )


class CompressedRLEType(pa.ExtensionType):
    """Segmentation mask type as PyArrow StructType"""

    def __init__(self):
        super(CompressedRLEType, self).__init__(
            pa.struct(
                [
                    pa.field("size", pa.list_(pa.int32())),
                    pa.field("counts", pa.binary()),
                ]
            ),
            "mask[rle]",
        )

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return CompressedRLEType()

    def __arrow_ext_serialize__(self):
        return b""


def is_image_type(t: pa.DataType) -> bool:
    """Returns True if value is an instance of ImageType

    Args:
        t (pa.DataType): Value to check

    Returns:
        bool: Type checking response
    """
    return isinstance(t, ImageType)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import datetime
import json
import os
from collections import defaultdict
from collections.abc import Iterator
from pathlib import Path
from urllib.parse import urlparse

import pyarrow as pa
import pyarrow.parquet as pq
from tqdm.auto import tqdm

from pixano.core import DatasetInfo, arrow_types
from pixano.transforms import (
    coco_names_91,
    denormalize,
    encode_rle,
    image_to_thumbnail,
    natural_key,
    normalize,
    rle_to_urle,
    urle_to_bbox,
)

from .data_loader import DataLoader


class COCOLoader(DataLoader):
    """Data Loader class for COCO instances dataset

    Attributes:
        name (str): Dataset name
        description (str): Dataset description
        splits (list[str]): Dataset splits
        schema (pa.schema): Dataset schema
        partitioning (ds.partitioning): Dataset partitioning
    """

    def __init__(
        self,
        name: str,
        description: str,
        splits: list[str],
    ):
        """Initialize COCO Loader

        Args:
            name (str): Dataset name
            description (str): Dataset description
            splits (list[str]): Dataset splits
        """

        # Dataset views
        views = [pa.field("image", arrow_types.ImageType())]

        # Initialize Data Loader
        super().__init__(name, description, splits, views)

    def import_row(
        self,
        input_dirs: dict[str, Path],
        split: str,
        portable: bool = False,
    ) -> Iterator:
        """Process dataset row for import

        Args:
            input_dirs (dict[str, Path]): Input directories
            split (str): Dataset split
            portable (bool, optional): True to move or download media files inside dataset. Defaults to False.

        Yields:
            Iterator: Processed rows
        """

        # Open annotation files
        with open(input_dirs["objects"] / f"instances_{split}.json", "r") as f:
            coco_instances = json.load(f)

        # Group annotations by image ID
        annotations = defaultdict(list)
        for ann in coco_instances["annotations"]:
            annotations[ann["image_id"]].append(ann)

        # Process rows
        for im in sorted(coco_instances["images"], key=lambda x: natural_key(x["id"])):
            # Load image annotations
            im_anns = annotations[im["id"]]
            # Load image
            file_name_uri = urlparse(im["file_name"])
            if file_name_uri.scheme == "":
                im_path = input_dirs["image"] / split / im["file_name"]
            else:
                im_path = Path(file_name_uri.path)

            # Create image thumbnail
            im_thumb = image_to_thumbnail(im_path.read_bytes())

            # Set image URI
            im_uri = (
                f"image/{split}/{im_path.name}"
                if portable
                else im_path.absolute().as_uri()
            )

            # Fill row with ID, image, and list of image annotations
            row = {
                "id": str(im["id"]),
                "image": arrow_types.Image(im_uri, None, im_thumb).to_dict(),
                "objects": [
                    arrow_types.ObjectAnnotation(
                        id=str(ann["id"]),
                        view_id="image",
                        area=float(ann["area"]) if ann["area"] else None,
                        bbox=normalize(ann["bbox"], im["height"], im["width"]),
                        mask=encode_rle(ann["segmentation"], im["height"], im["width"]),
                        is_group_of=bool(ann["iscrowd"]) if ann["iscrowd"] else None,
                        category_id=int(ann["category_id"]),
                        category_name=coco_names_91(ann["category_id"]),
                    ).dict()
                    for ann in im_anns
                ],
                "split": split,
            }

            # Return row
            yield row

    def export_dataset(
        self,
        input_dir: Path,
        export_dir: Path,
        input_dbs: list = [],
        portable: bool = False,
    ):
        """Export dataset back to original format

        Args:
            input_dir (Path): Input directory
            export_dir (Path): Export directory
            input_dbs (list[str]): Input databases to export, all if []. Defaults to [].
            portable (bool, optional): True to export dataset portable media files. Defaults to False.
        """

        # Load spec.json
        input_info = DatasetInfo.parse_file(input_dir / "spec.json")

        # Create URI prefix
        media_dir = input_dir / "media"
        uri_prefix = media_dir.absolute().as_uri()
        export_uri_prefix = (export_dir / "media").absolute().as_uri()

        # If no splits provided, select all splits
        if not self.splits:
            splits = [s.name for s in os.scandir(input_dir / "db") if s.is_dir()]
        # Else, format provided splits
        else:
            splits = [
                f"split={s}" if not s.startswith("split=") else s for s in self.splits
            ]
        # Check if the splits exist
        for split in splits:
            split_dir = input_dir / "db" / split
            if not split_dir.exists():
                raise Exception(f"{split_dir} does not exist.")
            if not any(split_dir.iterdir()):
                raise Exception(f"{split_dir} is empty.")

        # If no input databases provided, select all input databases
        if not input_dbs:
            input_dbs = ["db"]
            for inf_json in sorted(list(input_dir.glob("db_infer_*/infer.json"))):
                input_dbs.append(inf_json.parent.name)
        # Else, format provided inference datasets
        else:
            input_dbs = [
                f"db_infer_{i}" if not i.startswith("db") else i for i in input_dbs
            ]

        # Check if the datasets exist
        for ds in input_dbs:
            ds_dir = input_dir / ds
            if not ds_dir.exists():
                raise Exception(f"{ds_dir} does not exist.")
            if not any(ds_dir.iterdir()):
                raise Exception(f"{ds_dir} is empty.")

        # Create export directory
        ann_dir = export_dir / f"annotations_[{','.join(input_dbs)}]"
        ann_dir.mkdir(parents=True, exist_ok=True)

        # Iterate on splits
        for split in splits:
            # List split files
            files = (input_dir / "db" / split).glob("*.parquet")
            files = sorted(files, key=lambda x: natural_key(x.name))
            split_name = split.replace("split=", "")

            # Create COCO json
            coco_json = {
                "info": {
                    "description": input_info.name,
                    "url": "N/A",
                    "version": f"v{datetime.datetime.now().strftime('%y%m%d.%H%M%S')}",
                    "year": datetime.date.today().year,
                    "contributor": "Exported from Pixano",
                    "date_created": datetime.date.today().isoformat(),
                },
                "licences": [
                    {
                        "url": "N/A",
                        "id": 1,
                        "name": "Unknown",
                    },
                ],
                "images": [],
                "annotations": [],
                "categories": [],
            }

            # Iterate on files
            for file in tqdm(files, desc=f"Processing {split_name} split", position=0):
                seen_category_ids = [None]

                # Load media table
                media_fields = [
                    field.name
                    for field in self.schema
                    if arrow_types.is_image_type(field.type)
                ]
                media_table = pq.read_table(file).select(["id"] + media_fields)

                # Load annotation tables
                ann_files = [input_dir / ds / split / file.name for ds in input_dbs]
                ann_tables = [pq.read_table(f).select(["objects"]) for f in ann_files]

                # Iterate on rows
                for row in tqdm(
                    range(media_table.num_rows),
                    desc=f"Processing {file.name}",
                    position=1,
                ):
                    media_row = media_table.take([row])
                    ann_rows = [ann_table.take([row]) for ann_table in ann_tables]
                    images = {}

                    for field in media_fields:
                        # Open image
                        images[field] = media_row[field][0].as_py(uri_prefix)
                        im_uri = (
                            media_row[field][0].as_py(export_uri_prefix).uri
                            if portable
                            else images[field].uri
                        )
                        im_filename = Path(urlparse(im_uri).path).name
                        im_w, im_h = images[field].size
                        # Append image info
                        coco_json["images"].append(
                            {
                                "license": 1,
                                "coco_url": im_uri,
                                "file_name": im_filename,
                                "height": im_h,
                                "width": im_w,
                                "id": media_row["id"][0].as_py(),
                            }
                        )

                    for ann_row in ann_rows:
                        anns = ann_row["objects"][0].as_py()

                        for ann in anns:
                            # Append annotation
                            im_w, im_h = images[ann["view_id"]].size
                            urle = rle_to_urle(ann["mask"])
                            bbox = (
                                ann["bbox"]
                                if ann["bbox"] != [0.0, 0.0, 0.0, 0.0]
                                else urle_to_bbox(urle)
                            )
                            coco_json["annotations"].append(
                                {
                                    "segmentation": urle,
                                    "area": ann["area"],
                                    "iscrowd": 0,
                                    "image_id": media_row["id"][0].as_py(),
                                    "bbox": denormalize(bbox, im_h, im_w),
                                    "category_id": ann["category_id"],
                                    "category_name": ann["category_name"],
                                    "id": ann["id"],
                                }
                            )
                            # Append category if not seen yet
                            if (
                                ann["category_id"] not in seen_category_ids
                                and ann["category_name"] is not None
                            ):
                                coco_json["categories"].append(
                                    {
                                        "supercategory": "N/A",
                                        "id": ann["category_id"],
                                        "name": ann["category_name"],
                                    },
                                )
                                seen_category_ids.append(ann["category_id"])

            # Sort categories
            coco_json["categories"] = sorted(
                coco_json["categories"], key=lambda c: c["id"]
            )

            # Save COCO json
            with open(ann_dir / f"instances_{split_name}.json", "w") as f:
                json.dump(coco_json, f)

            # Move media directory if portable and directory exists
            if portable:
                if media_dir.exists():
                    media_dir.rename(export_dir / "media")
                else:
                    raise Exception(
                        f"Activated portable option for export but {media_dir} does not exist."
                    )



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
import os
from abc import ABC
from datetime import datetime
from io import BytesIO
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from tqdm.auto import tqdm

from pixano.core import arrow_types
from pixano.transforms import natural_key


class InferenceModel(ABC):
    """Abstract parent class for OfflineModel and OnlineModel

    Attributes:
        name (str): Model name
        id (str, optional): Model ID
        device (str, optional): Model GPU or CPU device
        source (str, optional): Model source
        info (str, optional): Additional model info
    """

    def __init__(
        self,
        name: str,
        id: str = "",
        device: str = "",
        source: str = "",
        info: str = "",
    ) -> None:
        """Initialize model name and ID

        Args:
            name (str): Model name
            id (str, optional): Model ID. Defaults to "".
            device (str, optional): Model GPU or CPU device. Defaults to "".
            source (str, optional): Model source. Defaults to "".
            info (str, optional): Additional model info. Defaults to "".
        """

        self.name = name
        if id == "":
            self.id = f"{datetime.now().strftime('%y%m%d_%H%M%S')}_{name}"
        else:
            self.id = id
        self.device = device
        self.source = source
        self.info = info

    def inference_batch(
        self,
        batch: pa.RecordBatch,
        view: str,
        uri_prefix: str,
        threshold: float = 0.0,
    ) -> list[list[arrow_types.ObjectAnnotation]]:
        """Inference preannotation for a batch

        Args:
            batch (pa.RecordBatch): Input batch
            view (str): Dataset view
            uri_prefix (str): URI prefix for media files
            threshold (float, optional): Confidence threshold. Defaults to 0.0.

        Returns:
            list[list[arrow_types.ObjectAnnotation]]: Model inferences as lists of ObjectAnnotation
        """

        pass

    def embedding_batch(
        self,
        batch: pa.RecordBatch,
        view: str,
        uri_prefix: str,
    ) -> list[np.ndarray]:
        """Embedding precomputing for a batch

        Args:
            batch (pa.RecordBatch): Input batch
            view (str): Dataset view
            uri_prefix (str): URI prefix for media files

        Returns:
            list[np.ndarray]: Model embeddings as NumPy arrays
        """

        pass

    def process_dataset(
        self,
        input_dir: Path,
        process_type: str,
        views: list[str],
        splits: list[str] = [],
        batch_size: int = 1,
        threshold: float = 0.0,
    ) -> Path:
        """Process dataset for inference preannotation or embedding precomputing

        Args:
            input_dir (Path): Input dataset directory
            process_type (str): Process type ('infer' for inference preannotation or 'embed' for embedding precomputing)
            views (list[str]): Dataset views
            splits (list[str], optional): Dataset splits, all if []. Defaults to [].
            batch_size (int, optional): Rows per batch. Defaults to 1.
            threshold (float, optional): Confidence threshold for predictions. Defaults to 0.0.

        Returns:
            Path: Output dataset directory
        """

        output_dir = input_dir / f"db_{process_type}_{self.id}"

        # Load spec.json
        with open(input_dir / "spec.json", "r") as f:
            spec_json = json.load(f)

        # Create URI prefix
        media_dir = input_dir / "media"
        uri_prefix = media_dir.absolute().as_uri()

        # If no splits provided, select all splits
        if not splits:
            splits = [s.name for s in os.scandir(input_dir / "db") if s.is_dir()]
        # Else, format provided splits
        else:
            splits = [f"split={s}" if not s.startswith("split=") else s for s in splits]
        # Check if the splits exist
        for split in splits:
            split_dir = input_dir / "db" / split
            if not split_dir.exists():
                raise Exception(f"{split_dir} does not exist.")
            if not any(split_dir.iterdir()):
                raise Exception(f"{split_dir} is empty.")

        # Create schema
        fields = [pa.field("id", pa.string())]
        if process_type == "infer":
            fields.extend(
                [
                    pa.field("objects", pa.list_(arrow_types.ObjectAnnotationType())),
                ]
            )
        elif process_type == "embed":
            fields.extend(
                [
                    pa.field(f"{view}_embedding", arrow_types.EmbeddingType())
                    for view in views
                ]
            )
        schema = pa.schema(fields)

        # Iterate on splits
        for split in splits:
            # List split files
            files = (input_dir / "db" / split).glob("*.parquet")
            files = sorted(files, key=lambda x: natural_key(x.name))

            # Create output split directory
            split_dir = output_dir / split
            split_dir.mkdir(parents=True, exist_ok=True)
            split_name = split.replace("split=", "")

            # Check for already processed files
            processed = [p.name for p in split_dir.glob("*.parquet")]

            # Iterate on files
            for file in tqdm(files, desc=f"Processing {split_name} split", position=0):
                # Process only remaining files
                if file.name not in processed:
                    # Load file into batches
                    table = pq.read_table(file)
                    batches = table.to_batches(max_chunksize=batch_size)

                    # Iterate on batches
                    data = {field.name: [] for field in schema}
                    for batch in tqdm(
                        batches, desc=f"Processing {file.name}", position=1
                    ):
                        # Add row IDs
                        data["id"].extend([str(row) for row in batch["id"]])
                        # For inferences
                        if process_type == "infer":
                            # Iterate on views
                            batch_inf = []
                            for view in views:
                                batch_inf.append(
                                    self.inference_batch(
                                        batch, view, uri_prefix, threshold
                                    )
                                )
                            # Regroup view inferences by row
                            data["objects"].append(
                                [
                                    inf.dict()
                                    for row in range(len(batch["id"]))
                                    for view_inf in batch_inf
                                    for inf in view_inf[row]
                                ]
                            )
                        # For embeddings
                        elif process_type == "embed":
                            # Iterate on views
                            for view in views:
                                view_emb = self.embedding_batch(batch, view, uri_prefix)
                                for emb in view_emb:
                                    emb_bytes = BytesIO()
                                    np.save(emb_bytes, emb)
                                    data[f"{view}_embedding"].append(
                                        emb_bytes.getvalue()
                                    )

                    # Convert ExtensionTypes
                    arrays = []
                    for field in schema:
                        arrays.append(
                            arrow_types.convert_field(
                                field_name=field.name,
                                field_type=field.type,
                                field_data=data[field.name],
                            )
                        )

                    # Save to file
                    pq.write_table(
                        pa.Table.from_arrays(arrays, schema=schema),
                        split_dir / file.name,
                    )

                    # Save.json
                    self.create_json(
                        output_dir=output_dir,
                        filename=process_type,
                        spec_json=spec_json,
                        num_elements=pq.read_metadata(split_dir / file.name).num_rows,
                    )

        return output_dir

    def create_json(
        self,
        output_dir: Path,
        filename: str,
        spec_json: dict,
        num_elements: int,
    ):
        """Save output .json

        Args:
            output_dir (Path): Output dataset directory
            filename (str): Output .json filename
            spec_json (dict): Input dataset .json
            num_elements (int): Number of processed rows
        """

        # Load existing .json
        if (output_dir / f"{filename}.json").is_file():
            with open(output_dir / f"{filename}.json", "r") as f:
                output_json = json.load(f)
            output_json["num_elements"] += num_elements

        # Or create .json from scratch
        else:
            output_json = {
                "id": spec_json["id"],
                "name": spec_json["name"],
                "description": spec_json["description"],
                "num_elements": num_elements,
                "model_id": self.id,
                "model_name": self.name,
                "model_source": self.source,
                "model_info": self.info,
            }

        # Save .json
        with open(output_dir / f"{filename}.json", "w") as f:
            json.dump(output_json, f)

    def export_to_onnx(self, library_dir: Path):
        """Export Torch model to ONNX

        Args:
            library_dir (Path): Dataset library directory
        """

        pass



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import numpy as np
from PIL import Image

from .image import rle_to_mask, urle_to_rle


def denormalize(coord: list[float], height: int, width: int) -> list[float]:
    """Denormalize coordinates

    Args:
        coord (list[float]): Normalized coordinates
        height (int): Height
        width (int): Width

    Returns:
        list[float]: Unnormalized coordinates
    """

    denorm = []

    for i, c in enumerate(coord):
        if i % 2 == 0:
            denorm.append(c * width)
        else:
            denorm.append(c * height)

    return denorm


def normalize(coord: list[float], height: int, width: int) -> list[float]:
    """Normalize coordinates

    Args:
        coord (list[float]): Unnormalized coordinates
        height (int): Height
        width (int): Width

    Returns:
        list[float]: Normalized coordinates
    """

    norm = []

    for i, c in enumerate(coord):
        if i % 2 == 0:
            norm.append(c / width)
        else:
            norm.append(c / height)

    return norm


def mask_to_bbox(mask: np.ndarray) -> list[float]:
    """Returns the smallest bounding box containing all the mask pixels

    Args:
        mask (np.ndarray): Mask as NumPy Array

    Returns:
        list[float]: Normalized xywh bounding box
    """

    height, width = mask.shape
    bool_mask = np.array(mask).astype(bool)

    # Find all columns and rows that contain ones
    rows = np.any(bool_mask, axis=1)
    cols = np.any(bool_mask, axis=0)

    # Find the min and max col/row index that contain ones
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    # Calculate bbox height and width
    w = (cmax - cmin + 1) / width
    h = (rmax - rmin + 1) / height

    return [cmin / width, rmin / height, w, h]


def urle_to_bbox(urle: dict) -> list[float]:
    """Returns the smallest bounding box containing all the mask pixels

    Args:
        urle (dict): Mask as uncompressed RLE

    Returns:
        list[float]: Normalized xywh bounding box
    """

    return mask_to_bbox(rle_to_mask(urle_to_rle(urle)))


def format_bbox(bbox, is_predicted=False, confidence=None) -> dict:
    """Convert bounding box to frontend format

    Args:
        bbox (list[float]): Bounding box
        is_predicted (bool, optional): True for prediction, False for ground truth. Defaults to False.
        confidence (float, optional): Bounding box confidence. Defaults to None.

    Returns:
        dict: Bounding box in frontend format
    """

    return {
        "x": float(bbox[0]),
        "y": float(bbox[1]),
        "width": float(bbox[2]),
        "height": float(bbox[3]),
        "is_predict": is_predicted,
        "confidence": confidence,
    }


def xywh_to_xyxy(xywh: list[float]) -> list[float]:
    """Convert bounding box coordinates from xywh (using top left point as reference) to xyxy

    Args:
        xywh (list[float]): xywh coordinates

    Returns:
        list[float]: xyxy coordinates
    """

    return [
        float(xywh[0]),
        float(xywh[1]),
        float(xywh[0] + xywh[2]),
        float(xywh[1] + xywh[3]),
    ]


def xyxy_to_xywh(xyxy: list[float]) -> list[float]:
    """Convert bounding box coordinates from xyxy to xywh (using top left point as reference)

    Args:
        xyxy (list[float]): xyxy coordinates

    Returns:
        list[float]: xywh coordinates
    """

    return [
        float(xyxy[0]),
        float(xyxy[1]),
        float(xyxy[2] - xyxy[0]),
        float(xyxy[3] - xyxy[1]),
    ]



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import base64
from io import BytesIO
from itertools import groupby

import cv2
import numpy as np
from PIL import Image
from pycocotools import mask as mask_api


def image_to_binary(image: Image.Image, format: str = "PNG") -> bytes:
    """Encode image from Pillow to binary

    Args:
        image (Image.Image): Image as Pillow
        format (str, optional): Image file extension. Defaults to "PNG".

    Returns:
        bytes: Image as binary
    """

    with BytesIO() as output_bytes:
        image.save(output_bytes, format)
        im_bytes = output_bytes.getvalue()

    return im_bytes


def image_to_thumbnail(image: bytes | Image.Image) -> bytes:
    """Generate image thumbnail

    Args:
        image (bytes | Image.Image): Image as binary or as Pillow

    Returns:
        bytes: Image thumbnail as binary
    """

    if isinstance(image, bytes):
        image = Image.open(BytesIO(image))

    image.thumbnail((128, 128))
    return image_to_binary(image)


def binary_to_url(im_bytes: bytes) -> str:
    """Encode image from binary to base 64 URL

    Args:
        im_bytes (bytes): Image as binary

    Returns:
        str: Image as base 64
    """

    if im_bytes is not None:
        encoded = base64.b64encode(im_bytes).decode("utf-8")
        return f"data:image;base64,{encoded}"
    else:
        return ""


def depth_file_to_binary(depth_path: str) -> bytes:
    """Encode depth file to RGB image in binary

    Args:
        depth_path (str): Depth file path

    Returns:
        bytes: Depth file as RGB image in binary
    """

    depth = cv2.imread(depth_path, cv2.IMREAD_ANYDEPTH).astype(np.float32)
    depth = depth_array_to_gray(depth)
    depth_rgb = Image.fromarray(depth)

    return image_to_binary(depth_rgb)


def depth_array_to_gray(
    depth: np.ndarray,
    valid_start: float = 0.2,
    valid_end: float = 1,
    scale: float = 1.0,
) -> np.ndarray:
    """Encode depth array to gray levels

    Args:
        depth (np.ndarray): Depth array
        valid_start (float, optional): Valid start. Defaults to 0.2.
        valid_end (float, optional): Valid end. Defaults to 1.
        scale (float, optional): Scale. Defaults to 1.0.

    Returns:
        np.ndarray: Depth array in gray levels
    """

    mask = depth > 1

    # Scale gives depth in mm
    depth_n = depth * scale * 0.001

    if mask.sum() > 0:
        depth_n[mask] -= depth_n[mask].min()
        depth_n[mask] /= depth_n[mask].max() / (valid_end - valid_start)
        depth_n[mask] += valid_start

    depth_n *= 255
    depth_n = cv2.applyColorMap(
        depth_n[:, :, np.newaxis].astype(np.uint8), cv2.COLORMAP_PLASMA
    )

    return depth_n


def encode_rle(mask: list[list] | dict, height: int, width: int) -> dict:
    """Encode mask from polygons / uncompressed RLE / RLE to RLE

    Args:
        mask (list[list] | dict): Mask as polygons / uncompressed RLE / RLE
        height (int): Image height
        width (int): Image width

    Returns:
        dict: Mask as RLE
    """

    if isinstance(mask, list):
        rle = polygons_to_rle(mask, height, width)
    elif isinstance(mask, dict):
        if isinstance(mask["counts"], list):
            rle = urle_to_rle(mask)
        else:
            rle = mask
    else:
        rle = None
    return rle


def mask_to_rle(mask: Image.Image) -> dict:
    """Encode mask from Pillow or NumPy array to RLE

    Args:
        mask (Image.Image): Mask as Pillow or NumPy array

    Returns:
        dict: Mask as RLE
    """

    if mask is not None:
        mask_array = np.asfortranarray(mask)
        return mask_api.encode(mask_array)


def rle_to_mask(rle: dict) -> np.ndarray:
    """Decode mask from RLE to NumPy array

    Args:
        rle (dict): Mask as RLE

    Returns:
        np.ndarray: Mask as NumPy array
    """

    if rle is not None:
        return mask_api.decode(rle)


def polygons_to_rle(polygons: list[list], height: int, width: int) -> dict:
    """Encode mask from polygons to RLE

    Args:
        polygons (list[list]): Mask as polygons
        height (int): Image height
        width (int): Image width

    Returns:
        dict: Mask as RLE
    """

    if polygons is not None:
        rles = mask_api.frPyObjects(polygons, height, width)
        return mask_api.merge(rles)


def rle_to_polygons(rle: dict) -> list[list]:
    """Encode mask from RLE to polygons

    Args:
        rle (dict): Mask as RLE

    Returns:
        list[list]: Mask as polygons
    """

    if rle is not None and "size" in rle:
        h, w = rle["size"]
        polygons, _ = mask_to_polygons(rle_to_mask(rle))

        # Normalize point coordinates
        for p in polygons:
            p[::2] /= w
            p[1::2] /= h

        # Cast to python list
        polygons = [p.tolist() for p in polygons]

        return polygons


def mask_to_polygons(mask: np.ndarray) -> tuple[list, bool]:
    """Encode mask from NumPy array to polygons

    Args:
        mask (np.ndarray): Mask as NumPy array

    Returns:
        list: Mask as polygons
        bool: Mask has holes
    """

    if mask is not None:
        # Some versions of cv2 does not support incontiguous arr
        mask = np.ascontiguousarray(mask)

        # cv2.RETR_CCOMP flag retrieves all the contours and arranges them to a 2-level hierarchy.
        # External contours (boundary) of the object are placed in hierarchy-1.
        # Internal contours (holes) are placed in hierarchy-2.
        # cv2.CHAIN_APPROX_NONE flag gets vertices of polygons from contours.
        res = cv2.findContours(
            mask.astype("uint8"), cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE
        )
        hierarchy = res[-1]

        # If mask is empty
        if hierarchy is None:
            return [], False

        # Check if mask has holes
        has_holes = (hierarchy.reshape(-1, 4)[:, 3] >= 0).sum() > 0

        res = res[-2]
        res = [x.flatten() for x in res]

        # The coordinates from OpenCV are integers in range [0, W-1 or H-1].
        # We add 0.5 to turn them into real-value coordinate space. A better solution
        # would be to first +0.5 and then dilate the returned polygon by 0.5.
        res = [x + 0.5 for x in res if len(x) >= 6]

        return res, has_holes


def urle_to_rle(urle: dict) -> dict:
    """Encode mask from uncompressed RLE to RLE

    Args:
        urle (dict): Mask as uncompressed RLE

    Returns:
        dict: Mask as RLE
    """

    if urle is not None:
        height, width = urle["size"]
        return mask_api.frPyObjects(urle, height, width)


def rle_to_urle(rle: dict) -> dict:
    """Encode mask from RLE to uncompressed RLE

    Args:
        rle (dict): Mask as RLE

    Returns:
        dict: Mask as uncompressed RLE
    """

    if rle is not None:
        mask = rle_to_mask(rle)
        urle = {"counts": [], "size": list(mask.shape)}
        counts = urle.get("counts")

        for i, (value, elements) in enumerate(groupby(mask.ravel(order="F"))):
            if i == 0 and value == 1:
                counts.append(0)
            counts.append(len(list(elements)))

        return urle



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
import random
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterator
from io import BytesIO
from itertools import islice
from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds
import shortuuid
from PIL import Image
from tqdm.auto import tqdm

from pixano.analytics import compute_stats
from pixano.core import DatasetInfo, arrow_types


def _batch_dict(iterator: Iterator, batch_size: int) -> Iterator:
    """Batch dicts

    Args:
        iterator (Iterator): Iterator
        batch_size (int): Batch size

    Yields:
        Iterator: Iterator in batches
    """

    it = iter(iterator)
    while batch := list(islice(it, batch_size)):
        batch_dict = defaultdict(list)
        for d in batch:
            for k, v in d.items():
                batch_dict[k].append(v)
        yield batch_dict


class DataLoader(ABC):
    """Abstract Data Loader class

    Attributes:
        name (str): Dataset name
        description (str): Dataset description
        splits (list[str]): Dataset splits
        schema (pa.schema): Dataset schema
        partitioning (ds.partitioning): Dataset partitioning
    """

    def __init__(
        self,
        name: str,
        description: str,
        splits: list[str],
        views: list[pa.field],
    ):
        """Initialize Data Loader

        Args:
            name (str): Dataset name
            description (str): Dataset description
            splits (list[str]): Dataset splits
            views (list[pa.field]): Dataset views
        """

        # Dataset info
        self.info = DatasetInfo(
            id=shortuuid.uuid(),
            name=name,
            description=description,
            num_elements=0,
            preview=None,
            categories=[],
        )
        self.splits = splits

        # Dataset schema
        fields = [
            pa.field("split", pa.string()),
            pa.field("id", pa.string()),
            pa.field("objects", pa.list_(arrow_types.ObjectAnnotationType())),
        ]
        fields.extend(views)
        self.schema = pa.schema(fields)
        self.partitioning = ds.partitioning(
            pa.schema([("split", pa.string())]), flavor="hive"
        )

    def create_json(self, import_dir: Path, categories: list[dict] = []):
        """Create dataset spec.json

        Args:
            import_dir (Path): Import directory
            categories (list[dict], optional): Dataset categories. Defaults to [].
        """

        with tqdm(desc="Creating dataset info file", total=1) as progress:
            # Read dataset
            dataset = ds.dataset(import_dir / "db", partitioning=self.partitioning)

            # Check number of rows in the created dataset
            self.info.num_elements = dataset.count_rows()

            # Add categories
            if categories:
                self.info.categories = categories

            # Create spec.json
            with open(import_dir / "spec.json", "w") as f:
                json.dump(vars(self.info), f)
            progress.update(1)

    def create_preview(self, import_dir: Path):
        """Create dataset preview image

        Args:
            import_dir (Path): Import directory
        """

        # Read dataset
        dataset = ds.dataset(import_dir / "db", partitioning=self.partitioning)

        # Get list of image fields
        image_fields = []
        for field in self.schema:
            if arrow_types.is_image_type(field.type):
                image_fields.append(field.name)

        if image_fields:
            with tqdm(desc="Creating dataset thumbnail", total=1) as progress:
                tile_w = 64
                tile_h = 64
                preview = Image.new("RGB", (3 * tile_w, 2 * tile_h))
                for i in range(6):
                    field = image_fields[i % len(image_fields)]
                    row_number = random.randrange(dataset.count_rows())
                    row = dataset.take([row_number]).to_pylist()[0]
                    with Image.open(BytesIO(row[field]._preview_bytes)) as im:
                        preview.paste(im, ((i % 3) * tile_w, (int(i / 3) % 2) * tile_h))
                preview.save(import_dir / "preview.png")
                progress.update(1)

    def create_stats(self, import_dir: Path):
        """Create dataset statistics

        Args:
            import_dir (Path): Import directory
        """

        # Reset json file
        open(import_dir / "db_feature_statistics.json", "w").close()
        # Create objects stats
        self.objects_stats(import_dir)
        # Create image stats
        self.image_stats(import_dir)

    def objects_stats(self, import_dir: Path):
        """Create dataset objects statistics

        Args:
            import_dir (Path): Import directory
        """

        # Read dataset
        dataset = ds.dataset(import_dir / "db", partitioning=self.partitioning)

        # Create stats if objects field exist
        objects = pa.field("objects", pa.list_(arrow_types.ObjectAnnotationType()))
        if objects in self.schema:
            # Create dataframe
            df = dataset.to_table(columns=["split", "objects"]).to_pandas()
            # Split objects in one object per row
            df = df.explode("objects")
            # Remove empty objects
            df = df[df["objects"].notnull()]

            # Get features
            features = []
            for split, object in tqdm(
                zip(df["split"], df["objects"]),
                desc="Computing objects stats",
                total=len(df.index),
            ):
                try:
                    area = 100 * (object["area"] / np.prod(object["mask"]["size"]))
                except TypeError:
                    area = None
                features.append(
                    {
                        "id": object["id"],
                        "view id": object["view_id"],
                        "objects - is group of": object["is_group_of"],
                        "objects - area (%)": area,
                        "objects - category": object["category_name"],
                        "split": split,
                    }
                )
            features_df = pd.DataFrame.from_records(features).astype(
                {
                    "id": "string",
                    "view id": "string",
                    "objects - is group of": bool,
                    "objects - area (%)": float,
                    "objects - category": "string",
                    "split": "string",
                }
            )

            # Initialize stats
            stats = [
                {
                    "name": "objects - category",
                    "type": "categorical",
                    "histogram": [],
                },
                {
                    "name": "objects - is group of",
                    "type": "categorical",
                    "histogram": [],
                },
                {
                    "name": "objects - area (%)",
                    "type": "numerical",
                    "range": [0.0, 100.0],
                    "histogram": [],
                },
            ]

            # Save stats
            self.save_stats(import_dir, stats, features_df)

    def image_stats(self, import_dir: Path):
        """Create dataset image statistics

        Args:
            import_dir (Path): Import directory
        """

        # Read dataset
        dataset = ds.dataset(import_dir / "db", partitioning=self.partitioning)

        # Create URI prefix
        media_dir = import_dir / "media"
        uri_prefix = media_dir.absolute().as_uri()

        # Iterate over dataset columns
        for field in dataset.schema:
            # If column has images
            if arrow_types.is_image_type(field.type):
                features = []
                rows = dataset.to_batches(columns=[field.name, "split"], batch_size=1)

                # Get features
                for row in tqdm(
                    rows,
                    desc=f"Computing {field.name} stats",
                    total=dataset.count_rows(),
                ):
                    # Open image
                    im = row[field.name][0].as_py(uri_prefix)
                    im_w, im_h = im.size
                    # Compute image features
                    aspect_ratio = round(im_w / im_h, 1)
                    features.append(
                        {
                            f"{field.name} - aspect ratio": aspect_ratio,
                            "split": row["split"][0].as_py(),
                        }
                    )
                features_df = pd.DataFrame.from_records(features).astype(
                    {
                        f"{field.name} - aspect ratio": "float",
                        "split": "string",
                    }
                )

                # Initialize stats
                stats = [
                    {
                        "name": f"{field.name} - aspect ratio",
                        "type": "numerical",
                        "histogram": [],
                        "range": [
                            features_df[f"{field.name} - aspect ratio"].min(),
                            features_df[f"{field.name} - aspect ratio"].max(),
                        ],
                    },
                ]

                # Save stats
                self.save_stats(import_dir, stats, features_df)

    def save_stats(self, import_dir: Path, stats: list[dict], df: pd.DataFrame):
        """Compute and save stats to json

        Args:
            import_dir (Path): Import directory
            stats (list[dict]): List of stats to save
            df (pd.DataFrame): DataFrame to create stats from
        """

        # Compute stats
        for split in self.splits:
            for stat in stats:
                split_df = df[df["split"] == split]
                stat["histogram"].extend(compute_stats(split_df, split, stat))

        # Check for existing db_feature_statistics.json
        with open(import_dir / "db_feature_statistics.json", "r") as f:
            try:
                stat_json = json.load(f)
                stat_json.extend(stats)
            except ValueError:
                stat_json = stats

        # Add to db_feature_statistics.json
        with open(import_dir / "db_feature_statistics.json", "w") as f:
            json.dump(stat_json, f)

    @abstractmethod
    def import_row(
        self,
        input_dirs: dict[str, Path],
        split: str,
        portable: bool = False,
    ) -> Iterator:
        """Process dataset row for import

        Args:
            input_dirs (dict[str, Path]): Input directories
            split (str): Dataset split
            portable (bool, optional): True to move or download media files inside dataset. Defaults to False.

        Yields:
            Iterator: Processed rows
        """

        pass

    def import_dataset(
        self,
        input_dirs: dict[str, Path],
        import_dir: Path,
        portable: bool = False,
        batch_size: int = 2048,
    ):
        """Import dataset to Pixano format

        Args:
            input_dirs (dict[str, Path]): Input directories
            import_dir (Path): Import directory
            portable (int, optional): True to move or download files inside import directory. Defaults to False.
            batch_size (int, optional): Number of rows per file. Defaults to 2048.
        """

        # Check input directories
        for source_path in input_dirs.values():
            if not source_path.exists():
                raise Exception(f"{source_path} does not exist.")
            if not any(source_path.iterdir()):
                raise Exception(f"{source_path} is empty.")

        # Dataset categories
        categories = []
        seen_category_ids = [None]

        # Iterate on splits
        for split in self.splits:
            batches = _batch_dict(
                self.import_row(input_dirs, split, portable), batch_size
            )
            # Iterate on batches
            for i, batch in tqdm(enumerate(batches), desc=f"Converting {split} split"):
                # Append batch categories
                for field in self.schema:
                    # If column has annotations
                    # TODO: Change to checking type when ObjectAnnotationType is rebuilt
                    if field.name == "objects":
                        for row in batch[field.name]:
                            for ann in row:
                                if (
                                    ann["category_id"] not in seen_category_ids
                                    and ann["category_name"] is not None
                                ):
                                    categories.append(
                                        {
                                            "supercategory": "N/A",
                                            "id": ann["category_id"],
                                            "name": ann["category_name"],
                                        },
                                    )
                                    seen_category_ids.append(ann["category_id"])

                # Convert batch fields to PyArrow format
                arrays = []
                for field in self.schema:
                    arrays.append(
                        arrow_types.convert_field(
                            field_name=field.name,
                            field_type=field.type,
                            field_data=batch[field.name],
                        )
                    )
                # Save to file
                ds.write_dataset(
                    data=pa.Table.from_arrays(arrays, schema=self.schema),
                    base_dir=import_dir / "db",
                    basename_template=f"part-{{i}}-{i}.parquet",
                    format="parquet",
                    existing_data_behavior="overwrite_or_ignore",
                    partitioning=self.partitioning,
                )

        # Sort categories
        categories = sorted(categories, key=lambda c: c["id"])

        # Create spec.json
        self.create_json(import_dir, categories)

        # Create preview.png
        self.create_preview(import_dir)

        # Move media directories if portable
        if portable:
            for field in tqdm(self.schema, desc=f"Moving media directories"):
                if arrow_types.is_image_type(field.type):
                    field_dir = import_dir / "media" / field.name
                    field_dir.mkdir(parents=True, exist_ok=True)
                    input_dirs[field.name].rename(field_dir)

        # Create stats
        self.create_stats(import_dir)

    def export_dataset(self, input_dir: Path, export_dir: Path):
        """Export dataset back to original format

        Args:
            input_dir (Path): Input directory
            export_dir (Path): Export directory
        """

        pass



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
#
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
from pathlib import Path

from fastapi import Depends, FastAPI, HTTPException, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from fastapi_pagination import Page, Params
from fastapi_pagination.api import add_pagination
from pydantic import BaseSettings

from pixano.core import (
    Dataset,
    DatasetInfo,
    EmbeddingDataset,
    InferenceDataset,
    arrow_types,
)

from . import db_utils


class Settings(BaseSettings):
    """Dataset library settings

    Attributes:
        data_dir (Path): Dataset library directory
    """

    data_dir: Path = Path.cwd() / "library"


def load_library(settings: Settings) -> list[DatasetInfo]:
    """Load all dataset info files in library

    Args:
        settings (Settings): Dataset library settings

    Returns:
        list[DatasetInfo]: Dataset info files
    """

    infos = []
    for spec in sorted(settings.data_dir.glob("*/spec.json")):
        # Load dataset info
        info = DatasetInfo.parse_file(spec)
        # Load thumbnail
        preview_path = spec.parent / "preview.png"
        if preview_path.is_file():
            im = arrow_types.Image(uri=preview_path.absolute().as_uri())
            info.preview = im.url

        # Load categories
        info.categories = getattr(info, "categories", [])
        if info.categories is None:
            info.categories = []
        # Save dataset info
        infos.append(info)
    return infos


def load_dataset(ds_id: str, settings: Settings) -> Dataset:
    """Load dataset based on its ID

    Args:
        ds_id (str): Dataset ID
        settings (Settings): Dataset library

    Returns:
        Dataset: Dataset
    """

    for spec in settings.data_dir.glob("*/spec.json"):
        info = DatasetInfo.parse_file(spec)
        if ds_id == info.id:
            return Dataset(spec.parent)


def load_dataset_stats(ds_id: str, settings: Settings) -> dict:
    """Load dataset stats based on its ID

    Args:
        ds_id (str): Dataset ID
        settings (Settings): Dataset Library

    Returns:
        list[dict]: Dataset stats
    """

    ds = load_dataset(ds_id, settings)
    if ds is not None:
        stats_file = ds.path / "db_feature_statistics.json"
        if stats_file.is_file():
            with open(stats_file, "r") as f:
                return json.load(f)


def create_app(settings: Settings) -> FastAPI:
    """Run Pixano app

    Args:
        settings (Settings): Dataset Library

    Raises:
        HTTPException: 404, Dataset is not found
        HTTPException: 404, Dataset stats are not found
        HTTPException: 404, Dataset is not found
        HTTPException: 404, Dataset is not found
        HTTPException: 404, Embedding dataset is not found

    Returns:
        FastAPI: Explorer app
    """
    app = FastAPI()

    app.add_middleware(
        CORSMiddleware,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Check if library exists
    if not settings.data_dir.exists():
        raise FileNotFoundError(
            f"Dataset library '{settings.data_dir.absolute()}' not found"
        )

    # Create models folder
    model_dir = settings.data_dir / "models"
    model_dir.mkdir(exist_ok=True)
    app.mount("/models", StaticFiles(directory=model_dir), name="models")

    @app.get("/datasets", response_model=list[DatasetInfo])
    async def get_datasets_list():
        return load_library(settings)

    @app.get("/datasets/{ds_id}/items", response_model=Page[db_utils.Features])
    async def get_dataset_items(ds_id, params: Params = Depends()):
        # Load dataset
        ds = load_dataset(ds_id, settings)
        if ds is None:
            raise HTTPException(status_code=404, detail="Dataset not found")
        # Return dataset items
        res = db_utils.get_items(ds.load(), params)
        if res is None:
            raise HTTPException(status_code=404, detail="Data not found")
        else:
            return res

    @app.get("/datasets/{ds_id}/stats")
    async def get_dataset_stats(ds_id):
        # Load dataset stats
        stats = load_dataset_stats(ds_id, settings)
        if stats is None:
            raise HTTPException(status_code=404, detail="Stats not found")
        # Return dataset stats
        return stats

    @app.get("/datasets/{ds_id}/items/{item_id}")
    async def get_dataset_item_details(ds_id: str, item_id: str):
        # Load dataset
        ds = load_dataset(ds_id, settings)
        if ds is None:
            raise HTTPException(status_code=404, detail="Dataset not found")

        # Load inference datasets
        inf_datasets = []
        for inf_json in sorted(list(ds.path.glob("db_infer_*/infer.json"))):
            inf_datasets.append(InferenceDataset(inf_json.parent).load())

        # Return item details
        return db_utils.get_item_details(ds.load(), item_id, ds.media_dir, inf_datasets)

    @app.post("/datasets/{ds_id}/items/{item_id}/{view}/embedding")
    async def get_dataset_item_view_embedding(ds_id: str, item_id: str, view: str):
        # Load dataset
        ds = load_dataset(ds_id, settings)
        if ds is None:
            raise HTTPException(status_code=404, detail="Dataset not found")

        # Load embedding dataset (currently selecting latest one)
        emb_ds = None
        for emb_json in sorted(list(ds.path.glob("db_embed_*/embed.json"))):
            emb_ds = EmbeddingDataset(emb_json.parent).load()
        if emb_ds is None:
            raise HTTPException(status_code=404, detail="Embedding dataset not found")

        # Return item embedding
        return Response(content=db_utils.get_item_view_embedding(emb_ds, item_id, view))

    @app.post(
        "/datasets/{ds_id}/items/{item_id}/annotations",
        response_model=list[arrow_types.ObjectAnnotation],
    )
    async def post_dataset_item_annotations(
        ds_id: str,
        item_id: str,
        annotations: list[arrow_types.ObjectAnnotation],
    ):
        # Load dataset
        ds = load_dataset(ds_id, settings)
        if ds is None:
            raise HTTPException(status_code=404, detail="Dataset not found")

        # Update dataset annotations
        db_utils.update_annotations(ds.path, item_id, annotations)

        return Response()

    add_pagination(app)

    return app


def app():
    settings = Settings()
    return create_app(settings)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import asyncio
import html
import json
from pathlib import Path

import fastapi
import IPython.display
import shortuuid
import uvicorn
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from .main import Settings, create_app


def get_env() -> str:
    """Get current running environment

    Returns:
        str: Running environment
    """

    # If Google colab import succeeds
    try:
        import google.colab
        import IPython
    except ImportError:
        pass
    else:
        if IPython.get_ipython() is not None:
            return "colab"

    # If IPython import succeeds
    try:
        import IPython
    except ImportError:
        pass
    else:
        ipython = IPython.get_ipython()
        if ipython is not None and ipython.has_trait("kernel"):
            return "ipython"

    # Else
    return "none"


def display_colab(url: str, port: int, height: int):
    """Display a Pixano app inside a Google Colab

    Args:
        url (str): Pixano app URL
        port (int): Pixano app port
        height (int): Frame height
    """

    # Define frame template
    shell = """
        (() => {
            const url = new URL(%URL%);
            const port = %PORT%;
            if (port) {
                url.port = port;
            }
            const iframe = document.createElement('iframe');
            iframe.src = url;
            iframe.setAttribute('width', '100%');
            iframe.setAttribute('height', '%HEIGHT%');
            iframe.setAttribute('frameborder', 0);
            document.body.appendChild(iframe);
        })();
    """

    # Replace variables in template
    replacements = [
        ("%HEIGHT%", "%d" % height),
        ("%PORT%", "%d" % port),
        ("%URL%", json.dumps(url)),
    ]
    for k, v in replacements:
        shell = shell.replace(k, v)

    # Display frame
    script = IPython.display.Javascript(shell)
    IPython.display.display(script)


def display_ipython(url: str, port: int, height: int):
    """Display a Pixano app inside a Jupyter notebook

    Args:
        url (str): Pixano app URL
        port (int): Pixano app port
        height (int): Frame height
    """

    # Define frame template
    shell = """
        <iframe id="%HTML_ID%" width="100%" height="%HEIGHT%" frameborder="0">
        </iframe>
        <script>
            (function() {
                const frame = document.getElementById(%JSON_ID%);
                const url = new URL(%URL%, window.location);
                const port = %PORT%;
                if (port) {
                    url.port = port;
                }
                frame.src = url;
            })();
        </script>
    """

    # Replace variables in template
    frame_id = f"frame-{shortuuid.uuid()}"
    replacements = [
        ("%HTML_ID%", html.escape(frame_id, quote=True)),
        ("%JSON_ID%", json.dumps(frame_id)),
        ("%HEIGHT%", "%d" % height),
        ("%PORT%", "%d" % port),
        ("%URL%", json.dumps(url)),
    ]
    for k, v in replacements:
        shell = shell.replace(k, v)

    # Display frame
    iframe = IPython.display.HTML(shell)
    IPython.display.display(iframe)


def display_cli(url: str, port: int, height: int):
    """Display a Pixano app inside a command line interface

    Args:
        url (str): Pixano app URL
        port (int): Pixano app port
        height (int): Frame height
    """

    print(f"Please visit {url}:{port} in a web browser.")


class PixanoApp:
    """Pixano App

    Attributes:
        config (uvicorn.Config): App config
        server (uvicorn.Server): App server
        task_function (typing.Callable): Run task function for running environment
        display_function (typing.Callable): Display function for running environment
    """

    def __init__(
        self,
        library_dir: str,
        assets_path: str,
        template_path: str,
        host: str = "127.0.0.1",
        port: int = 8000,
    ):
        """Initialize and run Pixano app

        Args:
            library_dir (str): Dataset library directory
            host (str, optional): App host. Defaults to "127.0.0.1".
            port (int, optional): App port. Defaults to 8000.
        """

        # Create app
        templates = Jinja2Templates(directory=template_path)
        settings = Settings(data_dir=Path(library_dir))
        app = create_app(settings)

        @app.get("/", response_class=HTMLResponse)
        def main(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        app.mount("/assets", StaticFiles(directory=assets_path), name="assets")
        self.config = uvicorn.Config(app, host=host, port=port)
        self.server = uvicorn.Server(self.config)

        # Get environmennt
        self.task_function = {
            "colab": asyncio.get_event_loop().create_task,
            "ipython": asyncio.get_event_loop().create_task,
            "none": asyncio.run,
        }[get_env()]
        self.display_function = {
            "colab": display_colab,
            "ipython": display_ipython,
            "none": display_cli,
        }[get_env()]

        # Serve app
        self.task_function(self.server.serve())

    def display(self, height: int = 1000) -> None:
        """Display Pixano app

        Args:
            height (int, optional): Frame height. Defaults to 1000.
        """

        # Wait for app to be online
        while not self.server.started:
            self.task_function(asyncio.wait(0.1))

        # Display app
        for server in self.server.servers:
            for socket in server.sockets:
                address = socket.getsockname()
                self.display_function(
                    url=f"http://{address[0]}", port=address[1], height=height
                )



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import click
import pkg_resources

from pixano.apps import PixanoApp

LOGO = """
                             ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▒
                           ▒▓▒                                       ▓
              █▓          ▓▓                                         ▓
                        ▒▓▒   ▒▓▓▓▓▓ ▒▓   ▓▒▒▓▓▓▓▒     ▒▓▓▓▓▓▒       ▓
▓▓ ▓▓▓▓▓▓▒    ▓▒  ▒▓   ▓▒   ▒▓      ▓██   ██     ▓▒  ▒█▓     ▒█▓     ▓
███      ▓▓   █▒   ▓█      ▒█▒       ██   █▒     ▒█  ▓▒        █     ▓
██        █▒  █▒    ▒█▓     █▒       ██   █▒     ▒█  ▓▓        █     ▓
██        █▒  █▒    ▓█▓█     █▒     ▓██   █▒     ▒█   ██     ▓█▒     ▓
██▓▒    ▒▓▒   █▒   ▓▒  ▒█▓    ▒▓▓▓▓▓ ▒▓   ▓      ▒▓    ▒▓▓▓▓▓▒       ▓
██  ▒▓▓▒▒     ▒   ▒      ▒▓▒                                         ▓
██                         █▓                                        ▓
"""
ASSETS_PATH = pkg_resources.resource_filename("pixano", "apps/explorer/dist/assets")
TEMPLATE_PATH = pkg_resources.resource_filename("pixano", "apps/explorer/dist")


class ExplorerApp(PixanoApp):
    """Pixano Explorer App

    Attributes:
        config (uvicorn.Config): App config
        server (uvicorn.Server): App server
        task_function (typing.Callable): Run task function for running environment
        display_function (typing.Callable): Display function for running environment
    """

    def __init__(
        self,
        library_dir: str,
        host: str = "127.0.0.1",
        port: int = 0,
    ) -> None:
        """Initialize and run Pixano Explorer app

        Args:
            library_dir (str): Dataset library directory
            host (str, optional): App host. Defaults to "127.0.0.1".
            port (int, optional): App port. Defaults to 0.
        """

        super().__init__(library_dir, ASSETS_PATH, TEMPLATE_PATH, host, port)


@click.command(context_settings={"auto_envvar_prefix": "UVICORN"})
@click.argument(
    "library_dir",
    type=str,
)
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this host.",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=0,
    help="Bind socket to this port.",
    show_default=True,
)
def main(library_dir: str, host: str, port: int):
    """Launch Pixano Explorer

    LIBRARY_DIR: Dataset library directory
    """
    ExplorerApp(library_dir, host, port)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
from collections import defaultdict
from collections.abc import Iterator
from io import BytesIO
from math import isnan
from pathlib import Path
from urllib.parse import urlparse

import pyarrow as pa
from PIL import Image
from pycocotools import mask as mask_api

from pixano.core import arrow_types
from pixano.transforms import denormalize, image_to_thumbnail, xyxy_to_xywh

from .data_loader import DataLoader


class LegacyLoader(DataLoader):
    """Data Loader class for Pixano legacy format datasets

    Attributes:
        name (str): Dataset name
        description (str): Dataset description
        splits (list[str]): Dataset splits
        views (list[str]): Dataset views
        json_files (dict[str, str]): json file paths relative to workspace
    """

    def __init__(
        self,
        name: str,
        description: str,
        splits: list[str],
        views: list[str],
        json_files: dict[str, str],
    ):
        """Initialize Pixano Legacy Loader

        Args:
            name (str): Dataset name
            description (str): Dataset description
            splits (list[str]): Dataset splits
            views (list[str]): Dataset views
            json_files (dict[str, str]): json file paths relative to workspace
        """

        self.views = views
        self.json_files = json_files

        # Initialize Data Loader
        super().__init__(
            name,
            description,
            splits,
            [pa.field(view, arrow_types.ImageType()) for view in views],
        )

    def import_row(
        self,
        input_dirs: dict[str, Path],
        split: str,
        portable: bool = False,
    ) -> Iterator:
        """Process dataset row for import

        Args:
            input_dirs (dict[str, Path]): Input directory workspace for json files and images
            split (str): Dataset split
            portable (bool, optional): True to move or download media files inside dataset. Defaults to False.

        Yields:
            Iterator: Processed rows
        """

        category_ids = {}
        feats = defaultdict(list)
        for view in self.views:
            # Open annotation files
            with open(input_dirs["workspace"] / self.json_files[view], "r") as f:
                pix_json = json.load(f)

                # Group annotations by image ID (timestamp)
                annotations = defaultdict(list)
                for ann in pix_json["annotations"]:
                    annotations[str(ann["timestamp"])].append(ann)

                # Process rows
                for im in sorted(
                    pix_json["data"]["children"], key=lambda x: x["timestamp"]
                ):
                    # Load image
                    file_name_uri = urlparse(im["path"])
                    if file_name_uri.scheme == "":
                        im_path = input_dirs["workspace"] / im["path"]
                    else:
                        im_path = Path(file_name_uri.path)

                    image = Image.open(BytesIO(im_path.read_bytes()))
                    im_w = image.width
                    im_h = image.height
                    im_thumb = image_to_thumbnail(image)

                    feats[str(im["timestamp"])].append(
                        {
                            "viewId": view,
                            "width": im_w,
                            "height": im_h,
                            "im_thumb": im_thumb,
                            "im_uri": (
                                f"image/{split}/{im_path.name}"
                                if portable
                                else im_path.absolute().as_uri()
                            ),
                            "anns": annotations[str(im["timestamp"])],
                        }
                    )

        for timestamp in feats:
            # Fill row with ID, image
            row = {
                "id": timestamp,
                "objects": [],
                "split": split,
            }
            for f in feats[timestamp]:
                row[f["viewId"]] = arrow_types.Image(
                    f["im_uri"], None, f["im_thumb"]
                ).to_dict()

                # Fill row with list of image annotations
                for ann in f["anns"]:
                    # collect categories to build category ids
                    if ann["category"] not in category_ids:
                        category_ids[ann["category"]] = len(category_ids)

                    bbox = [0.0, 0.0, 0.0, 0.0]
                    mask = None

                    if "geometry" in ann:
                        if (
                            ann["geometry"]["type"] == "polygon"
                            and ann["geometry"]["vertices"]
                        ):
                            # Polygon
                            # we have normalized coords, we must denorm before making RLE
                            if not isnan(ann["geometry"]["vertices"][0]):
                                if len(ann["geometry"]["vertices"]) > 4:
                                    denorm = denormalize(
                                        ann["geometry"]["vertices"],
                                        f["height"],
                                        f["width"],
                                    )
                                    rles = mask_api.frPyObjects(
                                        [denorm], f["height"], f["width"]
                                    )
                                    mask = mask_api.merge(rles)
                                else:
                                    print(
                                        "Polygon with 2 or less points. Discarded\n",
                                        ann["geometry"],
                                    )
                        elif (
                            ann["geometry"]["type"] == "mpolygon"
                            and ann["geometry"]["mvertices"]
                        ):
                            # MultiPolygon
                            if not isnan(ann["geometry"]["mvertices"][0][0]):
                                denorm = [
                                    denormalize(poly, f["height"], f["width"])
                                    for poly in ann["geometry"]["mvertices"]
                                ]
                                rles = mask_api.frPyObjects(
                                    denorm, f["height"], f["width"]
                                )
                                mask = mask_api.merge(rles)
                        elif (
                            ann["geometry"]["type"] == "rectangle"
                            and ann["geometry"]["vertices"]
                        ):  # BBox
                            if not isnan(ann["geometry"]["vertices"][0]):
                                denorm = denormalize(
                                    [ann["geometry"]["vertices"]],
                                    f["height"],
                                    f["width"],
                                )
                                bbox = xyxy_to_xywh(denorm)
                        elif (
                            ann["geometry"]["type"] == "graph"
                            and ann["geometry"]["vertices"]
                        ):  # Keypoints
                            print("Keypoints are not implemented yet")
                        else:
                            # print('Unknown geometry', ann['geometry']['type'])  # log can be annoying if many...
                            pass
                    else:
                        # Ca peut etre un mask, ou 3d, trackink... etc.
                        print("No geometry?")

                    row["objects"].append(
                        arrow_types.ObjectAnnotation(
                            id=str(ann["id"]),
                            view_id=f["viewId"],
                            bbox=bbox,
                            mask=mask,
                            is_group_of=bool(ann["iscrowd"])
                            if "iscrowd" in ann
                            else None,
                            category_id=category_ids[ann["category"]],
                            category_name=ann["category"],
                        ).dict()
                    )

            # Return row
            yield row



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pathlib import Path
from typing import Any

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from fastapi_pagination.api import create_page, resolve_params
from fastapi_pagination.bases import AbstractPage, AbstractParams
from pydantic import BaseModel

from pixano import transforms
from pixano.core import arrow_types
from pixano.transforms import natural_key, urle_to_bbox, urle_to_rle


class Feature(BaseModel):
    """Feature

    Attributes:
        name (str): Feature name
        dtype (str): Feature dtype
        value (Any): Feature value
    """

    name: str
    dtype: str
    value: Any


Features = list[Feature]


def get_item_details(
    dataset: ds.Dataset,
    item_id: str,
    media_dir: Path,
    inf_datasets: list[ds.Dataset] = [],
) -> dict:
    """Get item details

    Args:
        dataset (ds.Dataset): Dataset
        item_id (str): Selected item ID
        media_dir (Path): Dataset media directory
        inf_datasets (list[ds.Dataset], optional): List of inference datasets. Defaults to [].

    Returns:
        dict: ImageDetails features for UI
    """

    # Get item
    scanner = dataset.scanner(filter=ds.field("id").isin([item_id]))
    item = scanner.to_table().to_pylist()[0]

    # Get item inference objects
    for inf_ds in inf_datasets:
        inf_scanner = inf_ds.scanner(filter=ds.field("id").isin([item_id]))
        inf_item = inf_scanner.to_table().to_pylist()[0]
        if inf_item is not None:
            item["objects"].extend(inf_item["objects"])

    # Create features
    features = {
        "id": item["id"],
        "filename": None,
        "width": None,
        "height": None,
        "categoryStats": [],
        "views": {},
    }

    # Category statistics
    cat_ids = [
        obj["category_id"] for obj in item["objects"] if obj["category_id"] is not None
    ]
    cat_names = [
        obj["category_name"]
        for obj in item["objects"]
        if obj["category_id"] is not None
    ]
    cat, index, count = np.unique(cat_ids, return_index=True, return_counts=True)
    # Add to features
    features["categoryStats"] = [
        {
            "id": int(cat[i]),
            "name": str(cat_names[index[i]]),
            "count": int(count[i]),
        }
        for i in range(len(cat))
        if cat[i] is not None
    ]

    # Views
    for field in dataset.schema:
        if arrow_types.is_image_type(field.type):
            # Image
            image = item[field.name]
            image.uri_prefix = media_dir.absolute().as_uri()

            # Objects IDs
            ids = [obj["id"] for obj in item["objects"] if obj["view_id"] == field.name]
            # Categories
            cats = [
                {"id": obj["category_id"], "name": obj["category_name"]}
                for obj in item["objects"]
                if obj["view_id"] == field.name
                and obj["category_id"] is not None
                and obj["category_name"] is not None
            ]
            # Bounding boxes
            boxes = [
                transforms.format_bbox(
                    obj["bbox"],
                    obj["bbox_confidence"] is not None,
                    obj["bbox_confidence"],
                )
                for obj in item["objects"]
                if obj["view_id"] == field.name and obj["bbox"] is not None
            ]
            # Masks
            masks = [
                transforms.rle_to_urle(obj["mask"])
                for obj in item["objects"]
                if obj["view_id"] == field.name and obj["mask"] is not None
            ]
            # Add to features
            features["views"][field.name] = {
                "image": image.url,
                "objects": {
                    "id": ids,
                    "category": cats,
                    "boundingBox": boxes,
                    "segmentation": masks,
                },
            }

    return features


def get_items(dataset: ds.Dataset, params: AbstractParams = None) -> AbstractPage:
    """Get items

    Args:
        dataset (pa.Dataset): Dataset
        params (AbstractParams, optional): FastAPI params for pagination. Defaults to None.

    Returns:
        AbstractPage: List of Features for UI (DatasetExplorer)
    """

    # Get page parameters
    params = resolve_params(params)
    raw_params = params.to_raw_params()
    total = dataset.count_rows()

    # Get page items
    start = raw_params.offset
    stop = min(raw_params.offset + raw_params.limit, total)
    if start >= stop:
        return None
    items_table = dataset.take(range(start, stop))

    def _create_features(row: list) -> list[Feature]:
        """Create features based on field types

        Args:
            row (list): Input row

        Returns:
            list[Feature]: Row as list of features
        """

        features = []

        # Iterate on fields
        for field in dataset.schema:
            # Number fields
            if arrow_types.is_number(field.type):
                features.append(
                    Feature(name=field.name, dtype="number", value=row[field.name])
                )
            # Image fields
            elif arrow_types.is_image_type(field.type):
                thumbnail = row[field.name].preview_url
                features.append(
                    Feature(name=field.name, dtype="image", value=thumbnail)
                )
            # String fields
            elif pa.types.is_string(field.type):
                features.append(
                    Feature(name=field.name, dtype="text", value=row[field.name])
                )

        return features

    # Create items features
    items = [_create_features(e) for e in items_table.to_pylist()]

    return create_page(items, total=total, params=params)


def get_item_view_embedding(emb_ds: ds.Dataset, item_id: str, view: str) -> bytes:
    """Get item embedding for a view

    Args:
        emb_ds (ds.Dataset): Embedding dataset
        item_id (str): Item ID
        view (str): Item embedding view

    Returns:
        bytes: Embedding in base 64
    """

    # Get item
    emb_scanner = emb_ds.scanner(filter=ds.field("id").isin([item_id]))
    emb_item = emb_scanner.to_table().to_pylist()[0]
    return emb_item[f"{view}_embedding"]


def update_annotations(
    dataset_dir: Path,
    item_id: str,
    annotations: list[arrow_types.ObjectAnnotation],
):
    """Update dataset annotations

    Args:
        dataset_dir (Path): Dataset directory
        item_id (str): Item ID
        annotations (list[arrow_types.ObjectAnnotation]): Item annotations
    """

    # Convert URLE to RLE and add bounding box
    # TODO: Find a fix for image datasets
    item_anns = [o.dict() for o in annotations]
    for ann in item_anns:
        ann["bbox"] = urle_to_bbox(ann["mask"])
        ann["bbox_source"] = ann["mask_source"]
        ann["mask"] = urle_to_rle(ann["mask"])

    # Dataset files
    files = (dataset_dir / "db").glob("**/*.parquet")
    files = sorted(files, key=lambda x: natural_key(x.name))

    # Iterate on dataset files
    for file in files:
        # Look for updated item
        table = pq.read_table(file)
        filter = table.filter(pc.field("id").isin([item_id]))
        item = filter.to_pylist()

        # If item found
        if item != []:
            # Read table without item
            updated_table = table.filter(~pc.field("id").isin([item_id])).to_pydict()

            # Add item with updated annotations
            item[0]["objects"] = item_anns
            for field in table.schema:
                updated_table[field.name].append(item[0][field.name])

            # Sort table fields according to IDs
            for field in table.schema:
                if field.name != "id":
                    updated_table[field.name] = [
                        x
                        for _, x in sorted(
                            zip(updated_table["id"], updated_table[field.name]),
                            key=lambda pair: natural_key(pair[0]),
                        )
                    ]
            # Sort table IDs
            updated_table["id"] = sorted(updated_table["id"], key=natural_key)

            # Convert ExtensionTypes
            arrays = []
            for field in table.schema:
                # Convert image types to dict before PyArrow conversion
                # TODO: find a better way
                if arrow_types.is_image_type(field.type):
                    updated_table[field.name] = [
                        i.to_dict() for i in updated_table[field.name]
                    ]
                arrays.append(
                    arrow_types.convert_field(
                        field_name=field.name,
                        field_type=field.type,
                        field_data=updated_table[field.name],
                    )
                )

            # Save updated table
            pq.write_table(
                pa.Table.from_arrays(arrays, schema=table.schema),
                file,
            )
            return



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Optional

import pyarrow as pa
from pydantic import BaseModel

from .image import CompressedRLEType


class BBoxType(pa.ExtensionType):
    """Bounding box type as PyArrow list of PyArrow float32"""

    def __init__(self):
        super(BBoxType, self).__init__(pa.list_(pa.float32(), list_size=4), "bbox")

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return BBoxType()

    def __arrow_ext_serialize__(self):
        return b""


class EmbeddingType(pa.ExtensionType):
    """Embedding type as PyArrow binary"""

    def __init__(self):
        super(EmbeddingType, self).__init__(pa.binary(), "embedding")

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return EmbeddingType()

    def __arrow_ext_serialize__(self):
        return b""


class Embedding(BaseModel):
    """Embedding class

    Attributes:
        embedding (bytes): Embedding as binary
    """

    embedding: bytes


class ObjectAnnotation(BaseModel):
    """ObjectAnnotation class to contain all annotation data

    Attributes:
        id (str): Annotation unique ID
        view_id (str, optional): View ID (e.g. 'image', 'cam_2')
        bbox (list[float], optional): Bounding box coordinates in xywh format (using top left point as reference)
        bbox_source (str, optional): Bounding box source
        bbox_confidence (float, optional): Bounding box confidence
        is_group_of (bool, optional): is_group_of
        is_difficult (bool, optional): is_difficult
        is_truncated (bool, optional): is_truncated
        mask (dict[str, bytes], optional): Mask
        mask_source (str, optional): Mask source
        area (float, optional): area
        pose (dict[str, list[float]], optional): Pose
        category_id (int, optional): Category ID
        category_name (str, optional): Category name
        identity (str, optional): Identity
    """

    # Object ID and View ID
    id: str
    view_id: Optional[str] = None
    # Bounding Box
    bbox: Optional[list[float]] = [0.0] * 4
    bbox_source: Optional[str] = None
    bbox_confidence: Optional[float] = None
    is_group_of: Optional[bool] = None
    is_difficult: Optional[bool] = None
    is_truncated: Optional[bool] = None
    # Mask
    mask: Optional[dict] = None
    mask_source: Optional[str] = None
    area: Optional[float] = None
    # 6D Poses
    pose: Optional[dict[str, list[float]]] = {
        "cam_R_m2c": [0.0] * 9,
        "cam_t_m2c": [0.0] * 3,
    }
    # Category
    category_id: Optional[int] = None
    category_name: Optional[str] = None
    identity: Optional[str] = None


def ObjectAnnotationType() -> pa.StructType:
    """ObjectAnnotation type as PyArrow StructType

    Returns:
        pa.StructType: ObjectAnnotation StructType
    """

    pose_schema = pa.struct(
        [
            pa.field("cam_R_m2c", pa.list_(pa.float64(), list_size=9)),
            pa.field("cam_t_m2c", pa.list_(pa.float64(), list_size=3)),
        ]
    )

    return pa.struct(
        [
            # Object ID and View ID
            pa.field("id", pa.string()),
            pa.field("view_id", pa.string(), nullable=True),
            # Bounding Box
            pa.field("bbox", BBoxType(), nullable=True),
            pa.field("bbox_source", pa.string(), nullable=True),
            pa.field("bbox_confidence", pa.float32(), nullable=True),
            pa.field("is_group_of", pa.bool_(), nullable=True),
            pa.field("is_difficult", pa.bool_(), nullable=True),
            pa.field("is_truncated", pa.bool_(), nullable=True),
            # Mask
            pa.field("mask", CompressedRLEType(), nullable=True),
            pa.field("mask_source", pa.string(), nullable=True),
            pa.field("area", pa.float32(), nullable=True),
            # 6D Poses
            pa.field("pose", pose_schema, nullable=True),
            # Category
            pa.field("category_id", pa.int32(), nullable=True),
            pa.field("category_name", pa.string(), nullable=True),
            pa.field("identity", pa.string(), nullable=True),
        ]
    )


def is_embedding_type(t: pa.DataType) -> bool:
    """Returns True if value is an instance of EmbeddingType

    Args:
        t (pa.DataType): Value to check

    Returns:
        bool: Type checking response
    """
    return isinstance(t, EmbeddingType)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import glob
from collections.abc import Iterator
from pathlib import Path

import pyarrow as pa
import shortuuid

from pixano.core import arrow_types
from pixano.transforms import image_to_thumbnail, natural_key

from .data_loader import DataLoader


class ImageLoader(DataLoader):
    """Data Loader class for demo datasets

    Attributes:
        name (str): Dataset name
        description (str): Dataset description
        splits (list[str]): Dataset splits
        schema (pa.schema): Dataset schema
        partitioning (ds.partitioning): Dataset partitioning
    """

    def __init__(
        self,
        name: str,
        description: str,
        splits: list[str],
    ):
        """Initialize COCO Loader

        Args:
            name (str): Dataset name
            description (str): Dataset description
            splits (list[str]): Dataset splits
        """

        # Dataset views
        views = [pa.field("image", arrow_types.ImageType())]

        # Initialize Data Loader
        super().__init__(name, description, splits, views)

    def import_row(
        self,
        input_dirs: dict[str, Path],
        split: str,
        portable: bool = False,
    ) -> Iterator:
        """Process dataset row for import

        Args:
            input_dirs (dict[str, Path]): Input directories
            split (str): Dataset split
            portable (bool, optional): True to move or download media files inside dataset. Defaults to False.

        Yields:
            Iterator: Processed rows
        """

        # Get images paths
        image_paths = []
        for type in ["*.png", "*.jpg", "*.jpeg"]:
            image_paths.extend(glob.glob(str(input_dirs["image"] / split / type)))
        image_paths = [Path(p) for p in sorted(image_paths, key=natural_key)]

        # Process rows
        for im_path in image_paths:
            # Create image thumbnail
            im_thumb = image_to_thumbnail(im_path.read_bytes())

            # Set image URI
            im_uri = (
                f"image/{split}/{im_path.name}"
                if portable
                else im_path.absolute().as_uri()
            )

            # Fill row with ID, image, and list of image annotations
            row = {
                "id": im_path.name,
                "image": arrow_types.Image(im_uri, None, im_thumb).to_dict(),
                # TODO: find a way to return an empty list
                "objects": [
                    arrow_types.ObjectAnnotation(
                        id=shortuuid.uuid(),
                        category_id=0,
                    ).dict()
                ],
                "split": split,
            }

            # Return row
            yield row



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from .coco_loader import COCOLoader
from .data_loader import DataLoader
from .dota_loader import DOTALoader
from .image_loader import ImageLoader
from .legacy_loader import LegacyLoader

__all__ = [
    "DataLoader",
    "COCOLoader",
    "DOTALoader",
    "ImageLoader",
    "LegacyLoader",
]



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
import tempfile
import unittest
from pathlib import Path

from pixano.core import Dataset

TARGET_INFO = {
    "name": "My dataset",
    "description": "Dataset from a great AI project",
    "features": [{"feat1": "int32"}, {"feat2": "image"}],
}


class TestDataset(unittest.TestCase):
    def setUp(self):
        self.tmpdir = Path(tempfile.mkdtemp())

    def test_from_spec(self):
        with open(self.tmpdir / "info.json", "w") as f:
            json.dump(TARGET_INFO, f)

        dataset = Dataset(self.tmpdir)

        self.assertEqual(TARGET_INFO["name"], dataset.info.name)
        self.assertEqual(TARGET_INFO["description"], dataset.info.description)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pathlib import Path
from typing import Optional

import pyarrow as pa
import pyarrow.dataset as ds
import pydantic


class DatasetInfo(pydantic.BaseModel):
    """DatasetInfo

    Attributes:
        id (str): Dataset ID
        name (str): Dataset name
        description (str): Dataset description
        num_elements (int): Number of elements in dataset
        preview (str, optional): Dataset preview
        categories (list[dict], optional): Dataset categories
    """

    id: str
    name: str
    description: str
    num_elements: int
    preview: Optional[str]
    categories: Optional[list[dict]]


class Dataset:
    """Dataset class

    Attributes:
        _path (Path): Dataset path
        _info (DatasetInfo): Dataset info
        _partitioning (ds.partitioning): Dataset partitioning
    """

    _partitioning: ds.partitioning = ds.partitioning(
        pa.schema([("split", pa.string())]), flavor="hive"
    )

    def __init__(self, path: Path):
        """Initialize dataset

        Args:
            path (Path): Dataset path
        """

        self._path = path
        self._info = DatasetInfo.parse_file(self._path / "spec.json")

    @property
    def info(self):
        return self._info

    @property
    def path(self):
        return self._path

    @property
    def media_dir(self):
        return self._path / "media"

    def load(self):
        return ds.dataset(self._path / "db", partitioning=self._partitioning)


class InferenceDataset(Dataset):
    """Inference Dataset

    Attributes:
        _path (Path): Dataset path
        _info (DatasetInfo): Dataset info
        _partitioning (ds.partitioning): Dataset partitioning
    """

    def __init__(self, path: Path):
        self._path = path
        self._info = DatasetInfo.parse_file(self._path / "infer.json")

    def load(self):
        return ds.dataset(
            self._path, partitioning=self._partitioning, ignore_prefixes=["infer.json"]
        )


class EmbeddingDataset(Dataset):
    """Embedding Dataset

    Attributes:
        _path (Path): Dataset path
        _info (DatasetInfo): Dataset info
        _partitioning (ds.partitioning): Dataset partitioning
    """

    def __init__(self, path: Path):
        self._path = path
        self._info = DatasetInfo.parse_file(self._path / "embed.json")

    def load(self):
        return ds.dataset(
            self._path, partitioning=self._partitioning, ignore_prefixes=["embed.json"]
        )
